#!/usr/bin/env python3

"""Initialize logger and hand off control to the controller module"""

from datetime import datetime
import logging
import os
import sys

from soutools import settings, controller, helpers

helpers.clear_screen()

def main():
    get_settings = settings.Settings()

    # Get logging information from settings
    file_log_level = get_settings.get_value('logging.file_log_level')
    file_log_path = get_settings.get_value('logging.file_log_path')

    # Setup log path and ensure that folders exist and if not, then create them
    home_dir = os.path.expanduser('~')
    log_path = os.path.join(home_dir, file_log_path)
    helpers.setup_folder_structure(log_path)
    log_file = f'{log_path}soutools_log__{datetime.now():%Y-%m-%d_%H-%M-%S}.log'

    # Setup a basic logger
    logger = logging.getLogger('soutools')
    logger.setLevel(logging.DEBUG)

    # Setup file handler and add it to the logger
    fh = logging.FileHandler(filename=log_file)
    # Validate log level is a proper instance
    file_level = getattr(logging, file_log_level.upper(), None)
    if isinstance(file_level, int):
        fh.setLevel(level = file_level)
    else:
        error = f'Invalid value "{file_log_level}" specified for the file_log_level'
        error = helpers.colorme(error, 'red')
        blue_text = helpers.colorme('file_log_level', 'blue')
        error += f'\n  Check {blue_text} in the settings.toml'
        error += '\n  Valid values are "debug", "info", "warning", "error", and "critical"'
        print()
        sys.exit('  ' + error)
    # create formatter and add it to the handler
    formatter = logging.Formatter('%(asctime)2s - %(name)19s - %(levelname)8s - %(message)s')
    fh.setFormatter(formatter)
    # add the handler to the logger
    logger.addHandler(fh)

    # create console handler and add it to the logger
    ch = logging.StreamHandler()
    ch.setLevel(level = 'WARNING')
    # create formatter and add it to the handler
    formatter = logging.Formatter('%(levelname)8s - %(message)s')
    ch.setFormatter(formatter)
    # add the handlers to the logger
    logger.addHandler(ch)
    
    logger.debug('Logger initialized, calling controller menu function')
    controller.main_menu()


if __name__ == '__main__':
    main()



# controller.py

"""Controlls the main logic of the program"""

import logging
import sys

from soutools import menu, model, helpers

logger = logging.getLogger(__name__)
dashboard = model.MerakiModel()
org_id = helpers.get_settings.get_value('default_org_id')
org_name = helpers.get_settings.get_value('default_org_name')


def select_organization():
    global org_id, org_name
    logger.debug('The "select_organization" function called')
    org_id, org_name = dashboard.select_organization()
    logger.debug(f'The selected organization ID is "{org_id}"')
    logger.info(f'The selected organization name is "{org_name}"')


def view_organization():
    logger.debug('The "view_organization" function called')
    logger.debug('Checking if organization name is set')
    if not org_name:
        print('\nYou have not selected an organization yet\n')
        logger.debug('Notified user that organization name is not set')
    else:
        highlight_orgname = helpers.colorme(org_name, 'blue')
        highlight_orgid = helpers.colorme(org_id, 'blue')
        print(f'\nThe "{highlight_orgname}" organization with ID {highlight_orgid} is currently selected\n')
        logger.debug(f'Notified user that organization name is set to {org_name}')
    input('Press [ENTER] to continue...')


def wireless_options():
    menu_title = helpers.menu_title
    wireless_menu = menu.Menu(
    menu_title, [
    ('Generate a report of all sites of type wireless', wireless_report),
    ('Generate a report of all sites with a particular SSID', wireless_report_with_ssid_number),
    ('Update radius servers for a particular SSID', update_radius_settings),
    ('Enable/Disable all sites with a particular SSID', enable_disable_ssid),
    ('Return to the main menu', main_menu),
    ('Exit', quit)])
    while True:
        wireless_menu.get_input()


def wireless_report():
    logger.debug('The "wireless_report" function called')
    if org_id:
        networks = dashboard.get_wireless_networks(org_id)
    else:
        print('No organization selected, please select organization first\n')
    path = helpers.get_settings.get_value('wireless_networks.output_file')
    helpers.writelines_to_file(path, networks)
    input('Press [ENTER] to continue...')


def wireless_report_with_ssid_number():
    logger.debug('The "wireless_with_ssid_x_report" function called')
    dashboard.find_ssid_number()
    input('Press [ENTER] to continue...')


def update_radius_settings():
    make_changes = 'n'
    logger.debug('The "update_radius_settings" function called')
    warning = helpers.colorme('WARNING', 'red')
    highlight_orgname = helpers.colorme(org_name, 'blue')
    print(f'{warning} - You are about to make changes to the {highlight_orgname} organizaiton')
    make_changes = input('Are you sure you want to continue [Y/N] ').upper()
    if 'Y' in make_changes:
        print()
        dashboard.update_radius_servers()
        input('Press [ENTER] to continue...')
    else:
        wireless_options()
    

def enable_disable_ssid():
    ssid = helpers.get_settings.get_value('networks_with_ssid.ssid')
    menu_title = helpers.menu_title
    main_menu = menu.Menu(
    menu_title, [
    (f'Enable ssid "{ssid}" in the "{org_name}" org', enable_sites_with_ssid),
    (f'Disable ssid "{ssid}" in the "{org_name}" org', disable_sites_with_ssid),
    ('Return to wireless options', wireless_options),
    ('Exit', quit)])
    while True:
        main_menu.get_input()


def enable_sites_with_ssid():
    make_changes = 'n'
    logger.debug('The "enable_sites_with_ssid" function called')
    warning = helpers.colorme('WARNING', 'red')
    highlight_orgname = helpers.colorme(org_name, 'blue')
    print(f'{warning} - You are about to make changes to the {highlight_orgname} organizaiton')
    make_changes = input('Are you sure you want to continue [Y/N] ').upper()
    if 'Y' in make_changes:
        print()
        dashboard.enable_disable_paticular_ssid(True)
        input('Press [ENTER] to continue...')
    else:
        enable_disable_ssid()


def disable_sites_with_ssid():
    make_changes = 'n'
    logger.debug('The "enable_sites_with_ssid" function called')
    warning = helpers.colorme('WARNING', 'red')
    highlight_orgname = helpers.colorme(org_name, 'blue')
    print(f'{warning} - You are about to make changes to the {highlight_orgname} organizaiton')
    make_changes = input('Are you sure you want to continue [Y/N] ').upper()
    if 'Y' in make_changes:
        print()
        dashboard.enable_disable_paticular_ssid(False)
        input('Press [ENTER] to continue...')
    else:
        enable_disable_ssid()


def quit():
    helpers.clear_screen()
    sys.exit()


def main_menu():
    menu_title = helpers.menu_title
    main_menu = menu.Menu(
    menu_title, [
    ('Select an organization', select_organization),
    ('View the selected organization', view_organization),
    ('Wireless options', wireless_options),
    ('Exit', quit)])
    while True:
        main_menu.get_input()



# helpers.py

"""This module is used to assists the other modules with functions like writting to file"""

import logging
import platform
import os
import sys

from soutools import settings

logger = logging.getLogger(__name__)
get_settings = settings.Settings()
home_dir = os.path.expanduser('~')


def setup_folder_structure(path):
    logger.debug('The "setup_folder_structure" function called')
    logger.debug(f'Checking the path {path} for existance and creating if not exists')
    temp_path = os.path.split(path)
    path = temp_path[0]
    if not os.path.exists(path):
        logger.debug(f'making directory structure for {path}')
        os.makedirs(path)


def writelines_to_file(path, text):
    logger.debug('The "writelines_to_file" function called')
    # Find users home directory and join with settings location
    location = os.path.join(home_dir, path)
    logger.debug(f'The path for writelines_to_file is {location}')
    setup_folder_structure(location)
    
    # Write text to given path
    try:
        with open(location, 'w') as file_data:
            file_data.writelines(text)
    except FileNotFoundError:
        logger.error(f'Could not open {location}')
        sys.exit()


def validate_integer_in_range(end_range):
    """Prompts user to select a number within a defined range

    Args:
        end_range (int): The upper bound number that can be selected

    Returns:
        int: The selected number
    """

    while True:
        try:
            selected = int(input('\n  Select an option >> '))
            assert selected in range(1, end_range)
        except ValueError:
            print("\tThat is not an integer!\n")
        except AssertionError:
            print(f'\n\tYou must enter a number between 0 and {end_range}')
        else:
            break
    return selected


def get_networks_list(path):
    filepath = os.path.join(home_dir, path)
    networks = []
    try:
        with open(filepath, 'r') as netdata:
            for line in netdata:
                if not line.isspace():
                    networks.append(line.strip())
        return networks
    except FileNotFoundError:
        logger.error(f'Could not open {path}')
        sys.exit()        


def format_radius():
    # Retrieve settings 
    radius = get_settings.get_value('radius_servers')
    accounting = get_settings.get_value('radius_accounting')
    radius_port = radius['port']
    radius_secret = radius['secret']
    accounting_port = accounting['port']
    accounting_secret = accounting['secret']
    # Initialize payloads
    radius_payload = []
    accounting_payload = []
    # Format Radius settings
    if radius_port and radius_secret:
        for server in radius['servers']:
            temp = {'host':server, 'port':radius_port, 'secret':radius_secret}
            radius_payload.append(temp)
    else:
        for server in radius['servers']:
            server_info = server.split(';')
            temp = {'host':server_info[0], 
                'port':int(server_info[1]), 
                'secret':server_info[2]}
            radius_payload.append(temp)
    # Format Accounting settings
    if accounting_port and accounting_secret:
        for server in accounting['servers']:
            temp = {'host':server, 'port':accounting_port, 'secret':accounting_secret}
            accounting_payload.append(temp)
    else:
        for server in accounting['servers']:
            server_info = server.split(';')
            temp = {'host':server_info[0], 
                'port':int(server_info[1]), 
                'secret':server_info[2]}
            accounting_payload.append(temp)
    # Return the formatted payloads
    return radius_payload, accounting_payload


def progress_bar(progress, total, width=40):
    char = chr(9632)
    logger.debug(f'progress is {progress}')
    logger.debug(f'total is {total}')
    if progress >= total:
        logger.debug('finished has been set to true')
        fill_char = colorme(char, 'green')
    else:
        fill_char = colorme(char, 'red')
    completed = int(width * (progress / total))
    bar = 'Progress: [' + fill_char * completed + '-' * (width - completed) + '] '
    percent_done = round(progress / total * 100, 1)
    bar += str(percent_done) + '% ' + str(progress) + '/' + str(total)
    return bar


def colorme(msg, color):
    if color == 'red':
        wrapper = '\033[91m'
    elif color == 'blue':
        wrapper = '\033[94m'
    elif color == 'green':
        wrapper = '\033[92m'
    else:
        # Defaults to white if invalid color is given
        wrapper = '\033[47m'
    return wrapper + msg + '\033[0m'


def clear_screen():
    if(platform.system().lower()=='windows'):
        cmd = 'cls'
    else:
        cmd = 'clear'
    os.system(cmd)
    

# https://patorjk.com/software/taag/ using formatting "standard"

menu_title = colorme("""
   ____        _   _ _____           _     
  / ___|  ___ | | | |_   _|__   ___ | |___ 
  \___ \ / _ \| | | | | |/ _ \ / _ \| / __|
   ___) | (_) | |_| | | | (_) | (_) | \__ \\
  |____/ \___/ \___/  |_|\___/ \___/|_|___/
""", 'red')



# model.py

"""This module handles the interface between the controller and the Meraki Dashboard"""

import logging
import os
import sys

import meraki

from soutools import helpers

logger = logging.getLogger(__name__)
logger.info('Loading settings for the dashboard instance')


class MerakiModel:
    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        logger.info('Instantiated an instance of the model.MerakiModel class')
        if helpers.get_settings.get_value('meraki.api_key'):
            api_key = helpers.get_settings.get_value('meraki.api_key')
        elif helpers.get_settings.get_value('meraki.api_key_environment_variable') in os.environ:
            api_key = os.environ.get(helpers.get_settings.get_value('meraki.api_key_environment_variable'))
        else:
            logger.error('API Key not found')
            sys.exit('ERROR: API Key not found')


        # Instantiate an instance of the Meraki dashboard
        self.dashboard = meraki.DashboardAPI(
            api_key=api_key,
            output_log=helpers.get_settings.get_value('meraki.output_log'), 
            print_console=helpers.get_settings.get_value('meraki.print_to_console'), 
            suppress_logging=helpers.get_settings.get_value('meraki.suppress_logging'),
            log_file_prefix=helpers.get_settings.get_value('meraki.log_file_prefix'),
            log_path=helpers.get_settings.get_value('meraki.log_path')
        )


    def select_organization(self):
        """Lists all organizations and prompts the user to select one

        Args:
            dashboard (obj): The Meraki dashboard instance

        Returns:
            str: the selected organization ID
            str: the selected organization name
        """
        logger.debug('The "select_organization" function called from the model')
        organizations = None
        try:
            logger.debug('Trying to get organizations from the Meraki dashboard')
            organizations = self.dashboard.organizations.getOrganizations()
            logger.debug(f'organizations is equal to {organizations}')
        except meraki.exceptions.APIError:
            sys.exit('Check network settings and verify API Key')
        organizations.sort(key=lambda x: x["name"])
        counter = 1
        print('  Organizations to choose from:\n')
        for organization in organizations:
            name = helpers.colorme(f"{str(counter)} - {organization['name']}", 'green')
            print(f'    {name}')
            counter += 1
        selected = helpers.validate_integer_in_range(counter)
        return (
            organizations[int(selected) - 1]['id'],
            organizations[int(selected) - 1]['name']
        )


    def get_wireless_networks(self, organization_id):
        logger.debug(f'The "get_wireless_networks" function called with org_id {organization_id}')
        networks = self.dashboard.organizations.getOrganizationNetworks(
        organization_id, total_pages='all')
        total = 0
        wireless = 0
        no_wireless = 0
        wireless_networks = []
        for network in networks:
            if 'wireless' in network['productTypes']:
                wireless_networks.append(network['id'] + ',' + network['name'] + '\n')
                logger.info(f'network {network["name"]}, was identified as having some type of wireless')
                wireless += 1
                total += 1
            else:
                logger.info(f'network {network["name"]}, was identified as not having some type of wireless')
                no_wireless += 1
        logger.info(f'INFO: There are {wireless} sites with some type of wireless device')
        logger.info(f'INFO: There were {no_wireless} sites with no wireless')
        
        return wireless_networks


    def find_ssid_number(self):
        logger.debug('The "wheres_ssid" function called')
        search_ssid = helpers.get_settings.get_value('networks_with_ssid.ssid')
        input_file = helpers.get_settings.get_value('networks_with_ssid.input_file')
        networks = helpers.get_networks_list(input_file)
        path = helpers.get_settings.get_value('networks_with_ssid.output_file')
        logger.debug(f'Searching for ssid "{search_ssid}"')
        logger.debug(f'Searching for networks found in "{input_file}"')
        logger.debug(f'Writting to path "{path}"')
        progress = found_count = not_found_count = 0 
        total = len(networks)
        logger.info(f'The total number of networks is {total}')
        ssid_sites = []
        print()
        for network in networks:
            bar = helpers.progress_bar(progress, total)
            print(bar, end='', flush=True)
            site = network.split(',')
            site_id = site[0]
            site_name = site[1]
            logger.debug(f'Calling Meraki dashboard to pull all SSIDs for network {site_id}')
            ssids = self.dashboard.wireless.getNetworkWirelessSsids(site_id)
            found_ssid = False
            for ssid in ssids:
                if (ssid['name'] == search_ssid):
                    found_ssid = True
                    newline = f'{site_id},{site_name},{ssid["name"]},{ssid["number"]},{ssid["enabled"]}\n'
                    ssid_sites.append(newline)
                    logger.info(f'Found SSID {search_ssid} for "{site_name}"')
                    found_count += 1
                    progress += 1
            if not found_ssid:
                logger.info(f'SSID {search_ssid} was not found for "{site_name}"')
                not_found_count += 1
                progress += 1
            print('\b' * len(bar), end='', flush=True)
        bar = helpers.progress_bar(progress, total)
        print(bar, end='', flush=True)
        print('\n')
        helpers.writelines_to_file(path, ssid_sites)
        logger.info(f'Sites with SSID {search_ssid}, found {found_count}')
        logger.info(f'Sites without SSID {search_ssid}, found {not_found_count}')


    def update_radius_servers(self):
        logger.debug('The "update_radius_servers" function called')
        input_file = helpers.get_settings.get_value('radius.input_file')
        networks = helpers.get_networks_list(input_file)
        radius, accounting = helpers.format_radius()
        logger.debug(f'Searching for networks "{networks}"')
        logger.debug(f'Searching for radius servers "{radius}"')
        logger.debug(f'Searching for accounting servers "{accounting}"')
        
        progress = 0
        total = len(networks)
        for network in networks:
            bar = helpers.progress_bar(progress, total)
            print(bar, end='', flush=True)
            site = network.split(',')
            net_id = site[0]
            net_name = site[1]
            ssid_num = site[3]
            logger.info(f'Updating radius settings for {net_name} and SSID number {ssid_num}')
            logger.debug(f'Calling Meraki dashboard with {net_id} {ssid_num} {radius} {accounting}')
            self.dashboard.wireless.updateNetworkWirelessSsid(
                networkId=net_id,
                number=ssid_num, 
                radiusServers=radius,
                radiusAccountingServers=accounting
            )
            progress += 1
            print('\b' * len(bar), end='', flush=True)
        logger.info(f'Updated radius settings on {progress} SSIDs')
        bar = helpers.progress_bar(progress, total)
        print(bar, end='', flush=True)
        print('\n')


    def enable_disable_paticular_ssid(self, ssid_enabled):
        logger.debug('The "enable_disable_paticular_ssid" function called')
        status = 'enabled' if ssid_enabled else 'disabled'
        input_file = helpers.get_settings.get_value('ssid_enabled_status.input_file')
        networks = helpers.get_networks_list(input_file)
        progress = 0 
        total = len(networks)
        print()
        for network in networks:
            bar = helpers.progress_bar(progress, total)
            print(bar, end='', flush=True)
            site = network.split(',')
            site_id = site[0]
            site_name = site[1]
            ssid_name = site[2]
            ssid_num = site[3]
            logger.info(f'Changing status for {site_name} ssid {ssid_name} to {status}')
            self.dashboard.wireless.updateNetworkWirelessSsid(
                networkId=site_id,
                number=ssid_num, 
                enabled=ssid_enabled
            )
            progress += 1
            print('\b' * len(bar), end='', flush=True)
        bar = helpers.progress_bar(progress, total)
        print(bar, end='', flush=True)
        print('\n')
